# Created October 3 2021
# Shell midden volume calculation
# Script by: Robert Gustas
# --------------------------------------------------

# Import libraries
import os, sys, arcpy, math, datetime, csv

# Set directory and GDB paths. Directory is folder which holds the geodatabase as well as other intermediate data tables.
directory = r"YOUR_DIRECTORY_PATH"
gdb = os.path.join(directory, "YOUR.GDB") # Must be a .gdb file

# Specify environment settings.
arcpy.env.overwriteOutput = True
arcpy.env.workspace = gdb

# Specify the input midden polygon feature class and the field which contains maximum depth "Depth_Field" and the field which contains borden values "borden_Num". Borden numbers should be recorded in the following format: FdPe-1.
in_Poly = os.path.join(gdb, "YOUR_FEATURE_CLASS")
Depth_Field = 'YOUR_FIELD' # This may be left blank however a pre-designated value (see line 227) will be used if no value is supplied. Depths should be recorded in meters as an integer, the sctipt will exit if any value is not an integer.
borden_Num = 'YOUR_FIELD'

# Set a prefix value for file names that can be changed to prevent overwriting files between runs.
prefix = "YOUR_VALUE"

# Functions
#----------------------------------------------------
# Define a function to create the midden line, bounding box, and centroid features.
def midden_Vector(in_Poly, borden, id):
    # Convert the midden polygons to line features.
    out_Line = "Line_{}_{}".format(id, borden)
    arcpy.FeatureToLine_management([in_Poly], out_Line)

    # Calculate the minimum bounding box around the line features, this shape is used to generate length, width, and orientation information that is used latter.
    bounding_Out = "Bounding_2D_{}_{}".format(id, borden)
    arcpy.MinimumBoundingGeometry_management(in_Poly, bounding_Out, "RECTANGLE_BY_WIDTH", "", "", "MBG_FIELDS")

    # Calculate the midden centroid. This is the semi-elipsoid origin and its x and y cordinates are used in volume calculations.
    out_Center = "Centeroid_{}_{}".format(id, borden)
    arcpy.FeatureToPoint_management(bounding_Out, out_Center, "CENTROID")

    # Gather outputs into a list and return these values.
    vector_Output = [out_Line, bounding_Out, out_Center]
    return vector_Output

# Define a function to calculate the midden RandomRaster, clip this layer, and convert it to a point feature.
def midden_Raster(out_Line, cell_Size, borden, id, extent):
    # Calculate a 1x1 m grid across the midden extent. The values in the grid don't matter, the cell size is whats important as this will dictate the resolution of the volume calculation.
    out_Random = "Random_{}_{}".format(id, borden)
    arcpy.management.CreateRandomRaster(gdb, out_Random, "NORMAL", extent, cell_Size)

    # Clip the grid to the midden size.
    out_Clip = "Clip_{}_{}".format(id, borden)
    outExtractByMask = arcpy.sa.ExtractByMask(out_Random, extent)
    outExtractByMask.save(out_Clip)

    # Convert the raster grid layer to a point layer. As long as the the cell size is 1 m each point will represent 1 m^2 and by solving for the height of the ellipsoid over the point and summing we can derive semi-ellipsid volume.
    out_Point = "Point_{}_{}".format(id, borden)
    arcpy.RasterToPoint_conversion(out_Clip, out_Point, "Value")

    # Gather outputs into a list and return these values.
    raster_Output = [out_Random, out_Clip, out_Point]
    return raster_Output

# Define a function to calculate midden volume on a per point basis
def midden_Ellipsoid(out_center, out_Point, bounding_Out, height, borden, id):
    # Calculate the x and y values of each distance point.
    arcpy.AddXY_management(out_center)
    # Calculate the x and y values of the site centroid.
    arcpy.AddXY_management(out_Point)

    # Assign the centroid x and y cordinates to variable.
    centroid_List = []
    with arcpy.da.SearchCursor(out_center, ["POINT_X", "POINT_Y"]) as cursor:
        for row in cursor:
            item = [row[0], row[1]]
            centroid_List.append(item)
    center_X = centroid_List[0][0]
    center_Y = centroid_List[0][1]

    # Add new fields to the midden point layer to hold the distance of each point to the centroid.
    new_X = "New_X"
    new_Y = "New_Y"
    arcpy.AddField_management(out_Point, new_X, "FLOAT")
    arcpy.AddField_management(out_Point, new_Y, "FLOAT")

    # Calculate the x and y distance of each point from the midden centroid.
    arcpy.CalculateField_management(out_Point, new_X, "!POINT_X! - {}".format(center_X))
    arcpy.CalculateField_management(out_Point, new_Y, "!POINT_Y! - {}".format(center_Y))

    # Derive length and width of the polygon from the attribute information of the minimum bounding rectangle feature class. Store these values in a list.
    xy_List = []
    with arcpy.da.SearchCursor(bounding_Out, ["MBG_Width", "MBG_Length", "MBG_Orientation"]) as cursor:
        for row in cursor:
            width = row[0]
            length = row[1]
            orientation = row[2]
            xy_List.append(width)
            xy_List.append(length)
            xy_List.append(orientation)

    # Specify values for the ellipsoid calculation
    width = xy_List[0] # Width as determined by the minimum bounding rectangle.
    length = xy_List[1] # Length as determined by the minimum bounding rectangle.
    a = width/2 # Width of midden from origin.
    b = length/2 # Length of midden from origin.
    c = height/2 # Half of the depth of the midden.
    angle = xy_List[2] # Angular orientation of the midden.
    theta = math.radians(angle) # Midden orientation in radians.

    # Add a field to hold z values before angular adjustment. "new_Z" isn't  needed to calculate the volume but can be useful for trouble shooting and so is retained here.
    new_Z = "New_Z"
    arcpy.AddField_management(out_Point, new_Z, "FLOAT")
    # Add a field to hold z values after angular adjustment.
    prime_Z = "Z_Prime"
    arcpy.AddField_management(out_Point, prime_Z, "FLOAT")

    # Calculate z values and add them to the attribute table of the point layer.
    with arcpy.da.UpdateCursor(out_Point, [new_X, new_Y, new_Z, "OBJECTID", prime_Z]) as cursor:
        for row in cursor:
            x = row[0]  # X distance of point in relation to the center
            y = row[1]  # Y distance of the point in relation to the center
            z = math.sqrt(abs(((1-(x**2/a**2)-(y**2/b**2))*c**2)))

            # Calculate x' and y'
            x_Prime = ((-1 * x) * math.cos(theta)) + (y * math.sin(theta))
            y_Prime = ((-1 * x) * math.sin(theta)) - (y * math.cos(theta))

            # Calculate the z' value. Distance from the xy-plane passing through the origin to the surface of the semi-ellipsoid.
            # z' represents the depth of the midden and is allowed to float vertically for the volume calculation.
            z_Prime = math.sqrt(abs(((1-(x_Prime**2/a**2)-(y_Prime**2/b**2))*c**2)))

            # Update attribute table row values.
            row[2] = z
            row[4] = z_Prime
            cursor.updateRow(row)

    # Write the ellipsoid center coordinate, length, and width to a table. These are used to create an ellipsoid line feature which is used to identify those point which fall outside the ellipsoid boundary.
    ellipsis_List = [center_X, center_Y, length, width, theta]
    ellipsis_Table = "{}_Ellipsis_Table.csv".format(borden)
    X_Field = "X_Field"
    Y_Field = "Y_Field"
    Major_Field = "Major_Field"
    Minor_Field = "Minor_Field"
    Azimuth = "Azimuth"
    out_CSV(ellipsis_List, [X_Field, Y_Field, Major_Field, Minor_Field, Azimuth], ellipsis_Table)

    # Ingest the table with the ellipsoid measurements and use it to create a polyline of the perimeter of the ellipsoid (ellipsis).
    table_In = os.path.join(directory, "Ellipsis_Tables", ellipsis_Table)
    out_Ellipsis = "Ellipsis_{}".format(borden)
    spatial_ref = arcpy.Describe(in_Poly).spatialReference
    arcpy.TableToEllipse_management(table_In, out_Ellipsis, X_Field, Y_Field, Major_Field, Minor_Field, "METERS", Azimuth, "RADS", "", spatial_ref)

    # Convert the ellipsis polyline to a polygon.
    ellipsis_Poly = "Ellipsis_Poly_{}".format(borden)
    arcpy.FeatureToPolygon_management(out_Ellipsis, ellipsis_Poly)

    # Select point features outside the ellipsis.
    ellipsis_Layer = "ellipsis_Layer"
    arcpy.MakeFeatureLayer_management(ellipsis_Poly, ellipsis_Layer)
    point_Layer = "point_Layer"
    arcpy.MakeFeatureLayer_management(out_Point, point_Layer)
    ellipsis_Select = arcpy.SelectLayerByLocation_management([point_Layer], "INTERSECT", ellipsis_Layer, "", "NEW_SELECTION")
    ellipsis_Switch = arcpy.SelectLayerByLocation_management([point_Layer], "INTERSECT", ellipsis_Layer, "", "SWITCH_SELECTION")

    # Replace Z_Prime values for the points outside the ellipsis with a zero (0) height.
    arcpy.CalculateField_management(ellipsis_Switch, prime_Z, "{}".format(0))
    arcpy.SelectLayerByAttribute_management(ellipsis_Switch, "CLEAR_SELECTION")

    # Convert points to a raster. This latter supplements the final point file allowing for raster representations of the depth of the ellipsoid.
    out_Raster = "Z_Height_{}_{}".format(id, borden)
    arcpy.FeatureToRaster_conversion(ellipsis_Switch, prime_Z, out_Raster, 1)

    # Iterate through all elevation points and add the "Z_Prime" values together to get the semi-ellipsoid midden volume.
    volume = 0
    with arcpy.da.SearchCursor(out_Point, [prime_Z]) as point:
        for elevation in point:
            volume += elevation[0]

    # Gather and return the outputs of this function.
    return [out_Raster, volume]

# Define a function to write output values to a CSV file.
def out_CSV(in_List, field_List, out_Name):
    outCSV = os.path.join(directory, "Ellipsis_Tables", out_Name)
    with open(outCSV, 'w', newline='') as file:
        writer = csv.writer(file)
        writer.writerow(field_List)
        for item in [in_List]:
            writer.writerow(item)

# Run analysis
#--------------------------------------------------------------
# This variable will advance by one after each run and is used to select midden polygon features based on OID value.
counter = 1

# Create a list of all OID values, this is the list that will be interated through to select midden polygon features.
identifers = []
with arcpy.da.SearchCursor(in_Poly, ['OID@']) as cursor:
    for row in cursor:
        identifers.append(row[0])

# Make a feature layer from midden polygon layer
out_Layer = "Poly_Layer"
arcpy.MakeFeatureLayer_management(in_Poly, out_Layer)

# Create empty lists to hold all the polygons and point feature classes that are created.
polygons = []
points = []

# Iterate through midden polygons and preform analysis on them.
for item in identifers:
    print("OID = {}".format(item))

    # Select polygons based on their OID values.
    query = "OBJECTID_1 = {}".format(item) # The name of the OID field may need to be changed, check your input file for the field name.
    arcpy.SelectLayerByAttribute_management(out_Layer, "NEW_SELECTION", query)
    out_Midden = "Midden_{}".format(item)

    # Save the selected polygon to a new feature class.
    arcpy.FeatureClassToFeatureClass_conversion(out_Layer, gdb, out_Midden)

    # Get the Borden number of the polygon.
    with arcpy.da.SearchCursor(out_Midden, [borden_Num, Depth_Field]) as cursor:
        for row in cursor:
            borden = row[0] # Assign the Borden number to a variable.
            midden_Depth = row[1] # Assign the midden depth value to a variable.
            average_Depth = 1.27 # Set the average midden depth to be used if none is supplied.
            if midden_Depth is None: # Check if value is None and if True use the "average_Depth" value.
                new_Depth = average_Depth
                print("Average Depth Used")
            elif type(midden_Depth) == int: # Check if the depth has a integer data type and convert to cm.
                new_Depth = int(midden_Depth) / 100
            elif midden_Depth.isnumeric() == True: # Check if the depth value has a numeric data type and convert to cm.
                new_Depth = int(midden_Depth) / 100
            elif midden_Depth == "NA":
                new_Depth = average_Depth
                print("Average Depth Used")
            else: # Exit the script if a non integer or numeric value is encountered.
                print("Unrecognised Value: {}".format(midden_Depth))
                sys.exit()

    # Print the depth of the current midden.
    print("Midden Depth = {} m".format(new_Depth))

    # Convert "-" in site numbers to "_" ArcGIS doesn't like "-"but will except "_".
    new_Borden = borden.replace("-", "_")
    print("Borden Number = {}".format(new_Borden))

    # Set the grid size. DON'T CHANGE UNLESS YOU REALLY KNOW WHAT YOUR DOING. It affects all the math down stream which assumes the calculations are being done on a 1x1 m grid.
    pixel_Size = 1

    # Preform midden volume calculations
    vector = midden_Vector(out_Midden, new_Borden, item)
    raster = midden_Raster(vector[0], pixel_Size, new_Borden, item, out_Midden)
    ellipsoid_Volume = midden_Ellipsoid(vector[2], raster[2], vector[1], new_Depth, new_Borden, item)

    # Add the volume measurement to the midden polygon
    new_Volume = ellipsoid_Volume[1]
    vol_Field = "Volume"
    arcpy.AddField_management(out_Midden, vol_Field, "FLOAT")
    arcpy.CalculateField_management(out_Midden, vol_Field, new_Volume)
    print("Estimated Volume = {} m\u00b3".format(new_Volume))

    # Add the individual polygon files to the previously created lists.
    polygons.append(out_Midden)
    points.append(raster[2])

    # Set a point after which the "For" loop should be exited
    if counter == 0:
        break
    # Advance the counter allowing the next
    counter += 1

# Aggregate results
# -----------------------------------------------
# Merge all midden polygons together.
poly_Merge = "{}_Polygon_Merge".format(prefix)
arcpy.Merge_management(polygons, poly_Merge)

# Merge all midden points together.
point_Merge = "{}_Point_Merge".format(prefix)
arcpy.Merge_management(points, point_Merge)

# Convert the points into a raster.
all_Raster = "{}_Depth_Raster".format(prefix)
prime_Z = "Z_Prime"
arcpy.FeatureToRaster_conversion(point_Merge, prime_Z, all_Raster, pixel_Size)
